/*
   ----------------------------------------------------------------------------
   HouseholdCrossTab.prg
   Sample code from The dBASE Reports Book, second edition
   Chapter 20
   Copyright  2016, Kenneth J. Mayer
   
   Program used to call the GetYear form, and then generate a 
   Cross-tab report from the household table.
   
   This may be used for other purposes, as needed. 
   ----------------------------------------------------------------------------
*/// Set up a database object to point to the data:
local oData
oData = new database()
oData.databaseName := "DBASEREPORTS"
oData.active := true

// Before we do anything else, we should empty
// the table used for the Cross Tab report:
try
   oData.emptyTable( "HouseholdCT" )
catch( exception e )
   // if we have an error emptying the table
   // we should let the user know, and then
   // stop processing. Normally any error
   // at this point will either be a typo
   // in the table name, or the table
   // is in use somewhere and we can't empty it.
   msgbox( e.code+" - "+e.message+chr(13)+;
           "Could not empty table, report cancelled!",;
           "Unexpected error!", 16 )
   // deactivate/release data object:
   oData.active := false
   release object oData
   oData = null
   // return from this program
   return
endtry

// set up data objects for the tables needed
local qHousehold, rHousehold, fHousehold
local qCT, rCT, fCT

// Household table:
qHousehold = new query()
qHousehold.database := oData
qHousehold.sql      := "select * from Household"
qHousehold.active   := true

// create two object shortcuts
rHousehold = qHousehold.rowset
fHousehold = rHousehold.fields

// set first index we need:
rHousehold.indexName := "Cat_Unique"

// Cross Tab table:
qCT = new query()
qCT.database := oData
qCT.sql      := "select * from HouseholdCT"
qCT.active   := true

// create two object shortcuts
rCT = qCT.rowset
fCT = rCT.fields

// don't allow null values in fields when adding rows:
rCT.autoNullFields := false

// ----------------- Actual Processing -----------------
// The first thing we need to do is get the categories, and
// add rows to the Cross Tab table - one for each category:
local cCat
rHousehold.first() // first row of unique index for Category
do while not rHousehold.endOfSet
   cCat = fHousehold["Category"].value.rightTrim()
   // add a row to the Cross Tab table for each category:
   rCT.beginAppend()
   fCT["Category"].value := cCat
   rCT.save()
   rHousehold.next() // next row
enddo

// Now that we have the Categories in the Cross-Tab table, we need
// to get the totals, averages and all that. However, we also need
// to get the *year* for the report. We could hard-code it into the
// program, but that would require changing the program each time you
// wanted to generate the report, which is not very user-friendly,
// or efficient. Instead, we need a form to allow the user to select
// the year. The following assumes a four digit year (SET CENTURY ON).
set procedure to CT_GetYear.wfm additive
local f, nYear
f = new CT_GetYearForm()
f.YearSB.rangeMin := 2015 // set the "lowest value acceptable"
f.YearSB.rangeMax := year( date() ) // set largest to current year
f.YearSB.value := year( date() ) // set default to current year
f.readModal()

// get the year that was displayed
// in the Spinbox control "YearSB":
nYear = f.YearSB.value

// release and close the form:
f.release()
release object f
f = null
close procedure CT_GetYear.wfm

// set to a different index, otherwise
// we will never see more than the six
// unique records we have in the table.
// We'll be using a combination of
// filter techniques to get the data
// we need.
rHousehold.indexName := "EntryDate"

// We have the year, we need to move on. We need
// to set a filter - this can be done in many
// ways, but as the book says, let's keep it simple.
// We know the year, we know each year starts at
// January 1, and ends at December 31, we can
// use the filter property for the household rowset.
//
// by using the setRange method of the rowset,
// we take advantage of the index "EntryDate",
// and can limit the data in a very fast way,
// this is faster than using a filter ...
// The date is being converted from a string
// to a date value (CTOD()):
local dStart, dEnd
dStart = ctod( "01/01/"+nYear )
dEnd = ctod( "12/31/"+nYear )

// set the range:
rHousehold.setRange( dStart, dEnd )

// Now this is the tricky part of what we need to do.
// We need to get the totals for each category
// for each month of the year, as well as averages
// for each category and totals for each category.
// As the number of categories could easily change
// from time to time that the report is run, we need
// to be flexible. Since we have the categories stored
// in the CT table, we will loop through that table
// to get the data for each category. Then it gets
// even more interesting.

// For a larger amount of data, or a more complex set
// of calculations, you might consider a progress form
// that would display, showing how far along in the
// table you are. Testing this with the test data,
// the process is fairly fast, and doesn't need a
// progress form.
local cCat, nMonth
rCT.first()
do while not rCT.endOfSet
   // we need to go through the data and get information
   // for each category - totals and such for each month.
   // The first thing to do is get the current Category:
   cCat = fCT["Category"].value.rightTrim()
   // Then we set a filter (on top of the setRange!)
   // on the Household table:
   rHousehold.filter = [Category=']+cCat+[']
   rHousehold.first()
   // next we want to loop through the Household table:
   do while not rHousehold.endOfSet
      // While looping through the data, now filtered for
      // the category in the CrossTab table, we need
      // to accumulate data for the months, and for the
      // total field:
      fCT["CatTotal"].value += fHousehold["Amount"].value
      // For each month it's a little trickier.
      // The HouseholdCT table has one field before
      // the months, so one relatively simple way to
      // handle this is to get the month of the
      // record in the Household table's EntryDate field,
      // and add one to it, and accumulate the value
      // in the field in HouseholdCT that relates to
      // that number. For example, if we are looking
      // at a month of January, the second field in the
      // HouseholdCT table is "Jan", which is the appropriate
      // field to add the data to:
      nMonth = month( fHousehold["EntryDate"].value )
      fCT[ nMonth+1 ].value += fHousehold["Amount"].value
      // save changes in this record:
      rCT.save()
      // next Household record:
      rHousehold.next()
   enddo
   // before moving on, let's calculate the average and
   // store it in the table - this is the category total
   // (for the year) divided by 12 (12 months in a year):
   fCT["MonAvg"].value := ( fCT["CatTotal"].value / 12 )
   rCT.save() // save this in the record
   // next CrossTab record
   rCT.next()
enddo

// Done looping through the Categories
// and done creating the data we need for
// the report.

// Cleanup - deactivate data objects:
qCT.active := false
release object qCT
qCT = null
qHousehold.active := false
release object qHousehold
qHousehold = null
oData.active := false
release object oData
oData = null

// This is the code to generate the report - once
// displayed, the user can print it or not ...:
local oRep
set procedure to HouseholdCrossTab.rep additive
oRep = new HouseholdCrossTabReport()
// display it in a modal preview screen:
do :dUFLP:Preview.wfm with oRep, true // modal
// clean up memory:
oRep.release()

// -  end of program HouseholdCrossTab.prg
